from openmmtorch import TorchForce
from ase.io import read
from simtk import openmm, unit
from simtk.openmm import app
from openmm.app.topology import Topology
from openmm import *
from openmm.app import *
import numpy as np
import time
from openmm.app import PDBFile
from openmmtools import cache, mcmc, multistate
from openmmtools.multistate import ReplicaExchangeSampler
from openmmtools.states import GlobalParameterState, SamplerState, ThermodynamicState, CompoundThermodynamicState
from openmm.unit import kelvin, bar
import logging
import copy
import torch

atoms = read('mobley_7378987_in_water_box_17A_lastframe_old.pdb')

# MM force
force = CustomBondForce('param_a*r+param_b*r^2')

'''
class ForceModule(torch.nn.Module):
    """A central harmonic force with a user-defined global scale parameter"""
    def forward(self, positions, boxvectors, param_a, param_b):
        """The forward method returns the energy computed from positions.

        Parameters
        ----------
        positions : torch.Tensor with shape (nparticles,3)
           positions[i,k] is the position (in nanometers) of spatial dimension k of particle i
        scale : torch.Scalar
           A scalar tensor defined by 'TorchForce.addGlobalParameter'.
           Here, it scales the contribution to the potential.
           Note that parameters are passed in the order defined by `TorchForce.addGlobalParameter`, not by name.

        Returns
        -------
        potential : torch.Scalar
           The potential energy (in kJ/mol)
        """
        boxsize = boxvectors.diag()
        periodicPositions = positions - torch.floor(positions/boxsize)*boxsize
        
        return param_a*torch.sum(periodicPositions**2) + param_b

module = torch.jit.script(ForceModule())
force = TorchForce(module)
'''
force.addGlobalParameter('param_a', 0.5)
force.addGlobalParameter('param_b', 0.5)
force.setUsesPeriodicBoundaryConditions(True)

# force.addEnergyParameterDerivative('scale')
pdbfile = PDBFile('mobley_7378987_in_water_box_17A_lastframe_old.pdb')

positions = pdbfile.getPositions(asNumpy=True)
topology = pdbfile.getTopology()

system = openmm.System()
for i in range(len(atoms.get_atomic_numbers())):
    system.addParticle(atoms.get_masses()[i])
    
    
box_vectors = np.diag([ 15.948 for i in range(3)]) * unit.angstroms
system.setDefaultPeriodicBoxVectors(*box_vectors)

# Remove MM constraints
while system.getNumConstraints() > 0:
  system.removeConstraint(0)

# Remove MM forces
while system.getNumForces() > 0:
  system.removeForce(0)

assert system.getNumConstraints() == 0
assert system.getNumForces() == 0

class LambdaState(GlobalParameterState):
    param_a = GlobalParameterState.GlobalParameter('param_a', standard_value=1.0)
    param_b = GlobalParameterState.GlobalParameter('param_b', standard_value=1.0)

    def set_rest_parameters(self, value_a, value_b):
        """Set all defined lambda parameters to the given value.

        The undefined parameters (i.e. those being set to None) remain undefined.

        Parameters
        ----------
        new_value : float
            The new value for all defined parameters.
        """
        lambda_functions = {'param_a': lambda value_a, value_b : value_a,
                 'param_b' : lambda value_a, value_b : value_b,
                 }

        for parameter_name in self._parameters:
            if self._parameters[parameter_name] is not None:
                new_value = lambda_functions[parameter_name](value_a, value_b)
                setattr(self, parameter_name, new_value)


# cv = openmm.CustomCVForce("")
# cv.addGlobalParameter("param_a", 1)
# cv.addGlobalParameter("param_b", 1)
# tempSystem = openmm.System()
# tempSystem.addForce(force)
# interactingVarNames = []
# for idx, force in enumerate(tempSystem.getForces()):
#     name = f"allForce{idx+1}"
#     cv.addCollectiveVariable(name, copy.deepcopy(force))
#     interactingVarNames.append(name)

# assert len(interactingVarNames) > 0 

# interactingSum = "+".join(interactingVarNames)

# cv.setEnergyFunction(
#     f"({interactingSum})"
# )

# system.addForce(cv)
system.addForce(force)

barostat = MonteCarloBarostat(1*bar, 298.15*kelvin)
system.addForce(barostat)

lambda_state = LambdaState(param_a=0.5, param_b=0.3)

thermostate = ThermodynamicState(system, temperature=298.15 * unit.kelvin)
compound_thermodynamic_state = CompoundThermodynamicState(thermostate, composable_states=[lambda_state])

# Create thermodynamic states
sampler_state =  SamplerState(pdbfile.positions, box_vectors=system.getDefaultPeriodicBoxVectors())
lambda_schedule_lj_1 = np.linspace(1, 0, 2)
lambda_schedule_nn_1 = np.linspace(0, 0, 2)
lambda_schedule_lj_2 = np.linspace(0, 1, 2)
lambda_schedule_nn_2 = np.linspace(0, 1, 2)
lambda_schedule_a = np.array([1, 2, 3])
lambda_schedule_b = np.array([4, 5, 6])
thermodynamic_state_list = []
sampler_state_list = []
for i in range(len(lambda_schedule_a)):
    # Create a thermodynamic state withs REST interactions scaled to the given temperature
    value_a = lambda_schedule_a[i]
    value_b = lambda_schedule_b[i]
    compound_thermodynamic_state_copy = copy.deepcopy(compound_thermodynamic_state)
    compound_thermodynamic_state_copy.set_rest_parameters(value_a, value_b)
    print('compound_thermodynamic_state_copy.param_a:', compound_thermodynamic_state_copy.param_a)
    print('compound_thermodynamic_state_copy.param_b:', compound_thermodynamic_state_copy.param_b)

    thermodynamic_state_list.append(compound_thermodynamic_state_copy)
    sampler_state_list.append(copy.deepcopy(sampler_state))

move = mcmc.LangevinDynamicsMove(timestep=0.5*unit.femtoseconds, n_steps=2) # each move is 1 ps
simulation = ReplicaExchangeSampler(mcmc_moves=move, number_of_iterations=4)

# Run repex
# reporter_file = f"test_{os.getpid()}.nc"
reporter_file = "test.nc"
reporter = multistate.MultiStateReporter(reporter_file, checkpoint_interval=1)
simulation.create(thermodynamic_states=thermodynamic_state_list,
                  sampler_states=sampler_state,
                  storage=reporter)
start_time = time.time()
simulation.run()
end_time = time.time()

print("Time to sample dynamics %.2f s" % (end_time - start_time))
